///////////////////////////////////////////////////////////////////////////
// ASVM: Active Support Vector Machine                                   //
// Copyright 2000 David R. Musicant and Olvi L. Mangasarian.             //
// This software is free for academic use only.                          //
// For commercial use, contact musicant@cs.wisc.edu.                     //
///////////////////////////////////////////////////////////////////////////
// AsvmTrain.cpp: This file contains code to do the SVM training.        //
///////////////////////////////////////////////////////////////////////////

#include "DataSource.H"
#include "Utilities.H"
#include "StandAlone.H"
#include "Version.H"
#include <set>
#include <algorithm>
#include <cfloat>

#define BASICTOL 1e-9
#define PERTURB_VALUE 0.001

double gNU, gPERTURB, gTOL;
char gSOURCEFILENAME[1000],gTESTFILENAME[1000],gDATATYPE[1000];
char gOUTPUTFILENAME[1000];
bool gHEADER, gCONVERGENCE, gNORMALIZE, gSETNU, gPROMPT;
int gMAXITER;

// Compute optimality condition and objective as necessary
template<class T>
void EvaluateSolution(MatrixTemplate<T> *H, Vector *u, double nu,
		      double &optCond, double &obj) {

  DenseVector *vectorN1 = new DenseVector(H->cols);
  DenseVector *vectorN2 = new DenseVector(H->cols);
  DenseVector *vectorM1 = new DenseVector(H->rows);
  DenseVector *vectorM2 = new DenseVector(H->rows);

  //norm(u-pl(u-M*u+e)  
  MultAtB(H,u,vectorN1);
  MultAB(H,vectorN1,vectorM1);
  u->Copy(vectorM2);
  vectorM2->MultiplyInPlace(1/nu);
  vectorM2->AddInPlace(vectorM1);
  vectorM2->AddInPlace(-1);
  u->Copy(vectorM1);
  vectorM1->SubtractInPlace(vectorM2);
  vectorM1->PlusFunctionInPlace();
  u->Copy(vectorM2);
  vectorM2->SubtractInPlace(vectorM1);
  optCond = sqrt(vectorM2->Square());


  MultAtB(H,u,vectorN2);
  // Compute Objective
  // Htu = H'*u;
  // objective = 1/2*u'*u/nu + 1/2*Htu'*Htu - u'*e;
  obj = 0.5*1/nu*u->Square() + 0.5*vectorN2->Square() - u->Sum();

  delete vectorN1;
  delete vectorN2;
  delete vectorM1;
  delete vectorM2;
}

// Compute optimality condition and objective as necessary
template<class T>
void EvaluateFace(MatrixTemplate<T> *H, Vector *u, set<int> basic, 
		  double nu, double &optCond, double &obj) {

  DenseVector *vectorN1 = new DenseVector(H->cols);
  int b = basic.size();
  DenseVector *vectorb1 = new DenseVector(b);
  DenseVector *vectorb2 = new DenseVector(b);

  //norm(u-pl(u-M*u+e))  (all on the basic set)
  MultAtB(H,basic,u,basic,vectorN1);
  MultAB(H,basic,vectorN1,vectorb1);
  u->Strip(vectorb2,basic);
  vectorb2->MultiplyInPlace(1/nu);
  vectorb2->AddInPlace(vectorb1);
  vectorb2->AddInPlace(-1);
  u->Strip(vectorb1,basic);
  vectorb1->SubtractInPlace(vectorb2);
  vectorb1->PlusFunctionInPlace();
  u->Strip(vectorb2,basic);
  vectorb2->SubtractInPlace(vectorb1);
  optCond = sqrt(vectorb2->Square());

  MultAtB(H,u,vectorN1);
  // Compute Objective
  // Htu = H'*u;
  // objective = 1/2*u'*u/nu + 1/2*Htu'*Htu - u'*e;
  obj = 0.5*1/nu*u->Square() + 0.5*vectorN1->Square() - u->Sum();

  delete vectorN1;
  delete vectorb1;
  delete vectorb2;
}

// Gradient projection subroutine, called only when convergence issues
// arise. This is based on ideas in:
//   Bertsekas, D. _Nonlinear Programming_. Athena Scientific,
//   Belmont, MA, 1995, p. 205
template<class T>
void ProjectedGradient(MatrixTemplate<T> *H, double nu, Vector *u,
		       set<int> basic) {
  
  int m = basic.size();
  int i;
  DenseVector *vectorN1 = new DenseVector(H->cols);
  DenseVector *vectorM1 = new DenseVector(m);
  DenseVector *vectorM2 = new DenseVector(m);
  DenseVector *vectorMu = new DenseVector(m);
  DenseVector *d = new DenseVector(m);
  
  // Calculate M*u - e
  MultAtB(H,basic,u,basic,vectorN1);
  MultAB(H,basic,vectorN1,vectorM1);
  u->Strip(vectorM2,basic);
  vectorM2->MultiplyInPlace(1/nu);
  vectorM2->AddInPlace(vectorM1);
  vectorM2->Copy(vectorMu); // need later
  vectorM2->AddInPlace(-1);
  // Calculate uBar = plusfun(u-(M*u-e))
  u->Strip(d,basic);
  d->SubtractInPlace(vectorM2);
  d->PlusFunctionInPlace();
  // Calculate d = uBar - u
  u->Strip(vectorM1,basic);
  d->SubtractInPlace(vectorM1);

  // Calculate lambda = (e'*d-d'*M*u)/(d'*M*d);
  double etd = d->Sum();
  double dtMu = d->ScalarProduct(vectorMu);
  //// Calculate M*d
  MultAtB(H,basic,d,vectorN1);
  MultAB(H,basic,vectorN1,vectorM1);
  d->Copy(vectorM2);
  vectorM2->MultiplyInPlace(1/nu);
  vectorM2->AddInPlace(vectorM1);
  double dtMd = d->ScalarProduct(vectorM2);
  double lambda = (etd-dtMu)/(dtMd);

  // Calculate uOut = u + lambda*d;
  DenseVector *uOut = vectorM1;
  u->Strip(uOut,basic);
  uOut->AddInPlace(d);
  
  // Look for problem indices. If any, update lambda.
  u->Strip(vectorM1,basic);
  for (i=0; i < m; i++)
    if ((*uOut)[i] < -BASICTOL) {
      double newlambda = (*vectorM1)[i]/-(*d)[i];
      if (newlambda < lambda) lambda = newlambda;
    }
  
  // Recalculate u
  d->MultiplyInPlace(lambda);
  uOut->AddInPlace(d);

  // Update locations of u with uOut
  set<int>::iterator iterI;
  for (i=0, iterI=basic.begin(); iterI!=basic.end(); i++, iterI++)
    u->Assign(*iterI,(*uOut)[i]);

  delete vectorN1;
  delete vectorM1;
  delete vectorM2;
  delete vectorMu;
  delete d;
}

// Find a feasible minimum
template<class T>
void FeasibleMin(MatrixTemplate<T> *H, double nu, Vector *u, set<int> basic) {
  
  int m = basic.size();
  int n = H->cols;
  int i;
  DenseVector *vectorN1 = new DenseVector(n);
  DenseVector *vectorN2 = new DenseVector(n);
  DenseVector *vectorM1 = new DenseVector(m);
  DenseVector *vectorM2 = new DenseVector(m);
  DenseVector *d = new DenseVector(m);
  DenseVector *e = new DenseVector(H->rows);
  DenseMatrixType<double> *matrixSmall = new DenseMatrixType<double>(n,n);

  for (i=0; i<H->rows; i++)
    e->Assign(i,1);

  // Compute (nu*(Im*e-H*inv(I/nu+Ht*H)*(H'*e)
  MultAtB(H,basic,e,basic,vectorN1);
  MultAtB(H,basic,H,basic,matrixSmall);
  matrixSmall->DiagAddInPlace(1/nu);
  matrixSmall->GaussianInverse();
  MultAB(matrixSmall,vectorN1,vectorN2);
  MultAB(H,basic,vectorN2,vectorM1);
  for (i=0; i<m; i++)
    vectorM2->Assign(i,1);
  vectorM2->SubtractInPlace(vectorM1);
  vectorM2->MultiplyInPlace(nu);
  // Compute d = -(u-nu*(Im*e-H*inv(I/nu+Ht*H)*(H'*e)
  u->Strip(vectorM1,basic);
  vectorM2->SubtractInPlace(vectorM1);
  vectorM2->Copy(d);
  double lambda = 1;
  
  // Calculate uOut = u + lambda*d;
  DenseVector *uOut = vectorM1;
  u->Strip(uOut,basic);
  uOut->AddInPlace(d);
  u->Strip(vectorM2,basic);

  // Look for problem indices. If any, update lambda.
  for (i=0; i < m; i++)
    if ((*uOut)[i] < -BASICTOL) {
      double newlambda = (*vectorM2)[i]/-(*d)[i];
      if (newlambda < lambda) lambda = newlambda;
    }

  // Recalculate u
  d->MultiplyInPlace(lambda);
  u->Strip(uOut,basic);
  uOut->AddInPlace(d);

  // Update locations of u with uOut
  set<int>::iterator iterI;
  for (i=0, iterI=basic.begin(); iterI!=basic.end(); i++, iterI++)
    u->Assign(*iterI,(*uOut)[i]);

  delete vectorN1;
  delete vectorN2;
  delete vectorM1;
  delete vectorM2;
  delete d;
  delete e;
  delete matrixSmall;
}

// Function to get input from user, in both environments.
// Strings for prompt and response should be declared outside of this
// function.
#ifndef MEXFILE
void GetInput(char *prompt, char *response) {
  printf("%s",prompt);
  scanf("%s",response);
  printf(response);
}
#else
void GetInput(char *prompt, char *response) {
  mxArray *lhs[1],*rhs[2];
  rhs[0] = mxCreateString(prompt);
  rhs[1] = mxCreateString("s");
  mexCallMATLAB(1,lhs,2,rhs,"input");
  int buflen = mxGetM(lhs[0])*mxGetN(lhs[0])*sizeof(mxChar)+1;
  int status = mxGetString(lhs[0],response,buflen);
  if (status)
    ThrowIt("GetInput: Failed to get string back from MATLAB.");
}
#endif

// Simple function to display output to screen
void ShowOutput(int numIter, double optCond, double obj,
		set<int> basic, char *convCase) {
  PRINT("%3d\t%15.6g\t%15.6g\t%10d\t%s\n",numIter,optCond,obj,basic.size(),
	  convCase);
  FLUSH;
}

// Do the actual training process
template<class T>
void AsvmTrain(MatrixTemplate<T> *H, double nu, double tol, int maxIter,
	       Vector *u, double &obj, double &time,
	       int &numIter, bool bCONVERGENCE, bool bPROMPT) {
 
 if (u->Size() != H->rows)
    ThrowIt("AsvmTrain: Size of u is not correct.");

  int i;
  CPUTime clock;
  clock.Reset();
  int m = H->rows;
  unsigned int n = H->cols;
  Vector *e = new DenseVector(m);
  for (i=0; i<m; i++)
    e->Assign(i,1);

  double objOld = DBL_MAX;
  obj = objOld / 10;
  // Keep track of smallest objective value. It's supposed to be
  // monotonic, but this can be violated with numerical problems.
  double minObj = obj;
  int minObjSameIters = 0;
  numIter = 0;

  Vector *vectorN1 = new DenseVector(n);
  Vector *vectorN2 = new DenseVector(n);
  Vector *vectorM1 = new DenseVector(m);
  Vector *vectorM2 = new DenseVector(m);
  Vector *gradf = new DenseVector(m);
  Vector *uOut = new DenseVector(m);
  Vector *uOld = new DenseVector(m);
  Vector *uMin = new DenseVector(m);
  Vector *uNew = new DenseVector(m);
  Vector *d = new DenseVector(m);
  Vector *vectorMu = new DenseVector(m);
  Vector *vectorMd = new DenseVector(m);
  Vector *vector1 = 0;
  DenseMatrixType<double> *matrixSmall = new DenseMatrixType<double>(n,n);
  DenseMatrixType<double> *invHb=0;
  DenseMatrixType<double> *matrix1=0, *matrix2=0, *matrix3=0, *matrix4=0,
    *matrix5=0;

  // Initialize u
  // vectorM2 = plusfun(nu*(Im*e-H*inv(I/nu+Ht*H)*(H'*e)))
  MultAtB(H,e,vectorN1);
  MultAtB(H,H,matrixSmall);
  matrixSmall->DiagAddInPlace(1/nu);
  matrixSmall->GaussianInverse();
  MultAB(matrixSmall,vectorN1,vectorN2);
  MultAB(H,vectorN2,vectorM1);
  for (i=0; i<m; i++)
    vectorM2->Assign(i,1);
  vectorM2->SubtractInPlace(vectorM1);
  vectorM2->MultiplyInPlace(nu);
  vectorM2->PlusFunctionInPlace();
  vectorM2->Copy(u);

  double optCond = 1e6;
  bool buildNewInverse = true;
  set<int> nonbasic;
  set<int> oldnonbasic;
  set<int> newnonbasic;
  set<int> indices;
  set<int> basic;
  set<int> newbasic;
  
  for (i=0; i < m; i++)
    indices.insert(i);

  // Display header for output
  PRINT(VERSION);
  PRINT("\n");
  PRINT("Iter         Optimality      Objective     Basic set size\n");
  FLUSH;

  char prompt[100];
  char response[100];
  strcpy(prompt,"Stop ASVM here (Y/N)? ");
  while (optCond>tol && numIter < maxIter) {
    numIter++;

    // Store prior iteration values
    objOld = obj;
    u->Copy(uOld);

    // Handle stuck across iterations
    if (obj < minObj - BASICTOL) {
      minObj = obj;
      minObjSameIters = 0;
    } else
      minObjSameIters++;
    if (minObjSameIters >= 5) {
      minObjSameIters = 0;
      if (!gNORMALIZE) {
	PRINT("Algorithm stuck. ");
	PRINT("Rerun using 'normalize' option.\n");
	if (bPROMPT) {
	  GetInput(prompt,response);
	  if (!strncmp(response,"Y",1) || !strncmp(response,"y",1))
	    maxIter = numIter;
	}
      } else {
	PRINT("Algorithm stuck. ");
	PRINT("Consider making nu smaller.\n");
	if (bPROMPT) {
	  GetInput(prompt,response);
	  if (!strncmp(response,"Y",1) || !strncmp(response,"y",1))
	    maxIter = numIter;
	}
      }
    }
  
    // Find basic and nonbasic u
    oldnonbasic = nonbasic;
    nonbasic = u->findTolNonBasic(BASICTOL);
    newnonbasic.clear();
    set_difference(nonbasic.begin(), nonbasic.end(), 
		   oldnonbasic.begin(), oldnonbasic.end(),
		   inserter(newnonbasic, newnonbasic.begin()));

    basic.clear();
    set_difference(indices.begin(), indices.end(), nonbasic.begin(),
		   nonbasic.end(), inserter(basic, basic.begin()));

    // Update u
    int b = basic.size();
    int nn = newnonbasic.size();
    u->Copy(uOut);

    // Incremental update of inverse only makes sense when the number of
    // changed points is significantly less than dimensionality of problem
    if (buildNewInverse || newnonbasic.size() > n/2) {
      // Compute inverse directly
      // invHb = inv(In/nu+Hb'*Hb)
      if (invHb) delete invHb;
      invHb = new DenseMatrixType<double>(n,n);
      buildNewInverse = false;
      MultAtB(H,basic,H,basic,invHb);
      invHb->DiagAddInPlace(1/nu);
      invHb->GaussianInverse();
    } else if (newnonbasic.size() > 0) {
      if (matrix1) delete matrix1;
      if (matrix2) delete matrix2;
      if (matrix3) delete matrix3;
      if (matrix4) delete matrix4;
      if (matrix5) delete matrix5;
      matrix1 = new DenseMatrixType<double>(nn,n);
      matrix2 = new DenseMatrixType<double>(nn,nn);
      matrix3 = new DenseMatrixType<double>(nn,n);
      matrix4 = new DenseMatrixType<double>(n,n);
      matrix5 = new DenseMatrixType<double>(n,n);
      MultAB(H,newnonbasic,invHb,matrix1);
      MultABt(matrix1,H,newnonbasic,matrix2);
      matrix2->MultiplyInPlace(-1);
      matrix2->DiagAddInPlace(1);
      matrix2->GaussianInverse();
      MultAB(matrix2,matrix1,matrix3);
      MultAtB(H,newnonbasic,matrix3,matrix4);
      MultAB(invHb,matrix4,matrix5);
      invHb->AddInPlace(matrix5);
    }

    // Update u
    // uOut(basic) = nu*(Ib*eb-Hb*(invHb*(Hb'*eb)));
    MultAtB(H,basic,e,basic,vectorN1);
    MultAB(invHb,vectorN1,vectorN2);
    if (vector1) delete vector1;
    vector1 = new DenseVector(b);
    MultAB(H,basic,vectorN2,vector1);
    vector1->SubtractInPlaceFrom(1);
    vector1->MultiplyInPlace(nu);
    uOut->Update(vector1,basic);
    uOut->Copy(uMin);
    uOut->PlusFunctionInPlace();
    uOut->Copy(u);

    EvaluateSolution(H,u,nu,optCond,obj);

    // Deal with convergence issues if not at final optimality
    char convCase[100];
    strcpy(convCase,"Normal iteration");
    if (obj >= objOld)
      if (!bCONVERGENCE) {
	PRINT("Objective unchanged. Rerun using 'converge' option.\n");
	if (bPROMPT) {
	  GetInput(prompt,response);
	  if (!strncmp(response,"Y",1) || !strncmp(response,"y",1))
	    maxIter = numIter;
	}
      } else {
	uOld->Copy(u);
	obj = objOld;
	basic = u->findTolBasic(BASICTOL);
	EvaluateFace(H,u,basic,nu,optCond,obj);
	if (optCond < tol) {
	  // If at optimality on the face, find all nonfeasible points.
	  // Use projected gradient to find new point. Basic must change.
	  ProjectedGradient(H,nu,u,indices);
	  strcpy(convCase,"Global grad projection");
	  buildNewInverse = TRUE;
	} else {
	  // Not at optimality on the face. Find global minimum to
	  // find a new point.
	  FeasibleMin(H,nu,u,basic);
	  EvaluateFace(H,u,basic,nu,optCond,obj);
	  strcpy(convCase,"Project active set min\0");
	  set<int> basicHere = u->findTolBasic(BASICTOL);
	  set<int> commonB;
	  set_intersection(basicHere.begin(), basicHere.end(), 
			   basic.begin(), basic.end(),
			   inserter(commonB, commonB.begin()));
	  // Check if improvement, if basic set changed, or at face
	  // optimality. If not, do projected gradient until face
	  // optimality or until face changes.
	  if (obj > objOld || (commonB.size() == basic.size() 
			       && optCond >= tol)) {
	    if ((obj > objOld + BASICTOL) && optCond >= tol) {
	      obj = objOld;
	      uOld->Copy(u);
	    }
	    int projIter = 0;
	    while (commonB.size() == basic.size() && optCond >= tol &&
		   numIter < maxIter) {
	      if (projIter >= 5) {
		projIter = 0;
		if (!gNORMALIZE) {
		  PRINT("Objective unchanged. ");
		  PRINT("Rerun using 'normalize' option.\n");
		  if (bPROMPT) {
		    GetInput(prompt,response);
		    if (!strncmp(response,"Y",1) || !strncmp(response,"y",1))
		      maxIter = numIter;
		  }
		} else {
		  PRINT("Objective unchanged. ");
		  PRINT("Consider making nu smaller.\n");
		  if (bPROMPT) {
		    GetInput(prompt,response);
		    if (!strncmp(response,"Y",1) || !strncmp(response,"y",1))
		      maxIter = numIter;
		  }
		}
	      }
	      ProjectedGradient(H,nu,u,basic);
	      basicHere = u->findTolBasic(BASICTOL);
	      set_intersection(basicHere.begin(), basicHere.end(), 
			       basic.begin(), basic.end(),
			       inserter(commonB, commonB.begin()));
	      EvaluateFace(H,u,basic,nu,optCond,obj);
	      strcpy(convCase,"Face grad loop");
	      if (projIter)
		numIter++;
	      projIter++;
	      double globalOptCond;
	      EvaluateSolution(H,u,nu,globalOptCond,obj);
	      ShowOutput(numIter,globalOptCond,obj,basic,convCase);
	      strcpy(convCase,"Face grad done");
	    }
	  }
	} // Optimality on face
      } // No progress on objective
    
    if (bCONVERGENCE) {
      EvaluateSolution(H,u,nu,optCond,obj);
      basic = u->findTolBasic(BASICTOL);
    }

    // Display output
    ShowOutput(numIter,optCond,obj,basic,convCase);
    
  } //end-while

  // Finished
  if (numIter==maxIter) {
    PRINT("Hit maximum number of iterations. Nu = %g.\n",nu);
    FLUSH;
  }

  delete e;
  delete vectorN1;
  delete vectorN2;
  delete vectorM1;
  delete vectorM2;
  delete gradf;
  delete uOut;
  delete uOld;
  delete uMin;
  delete uNew;
  delete d;
  delete vectorMu;
  delete vectorMd;
  delete matrixSmall;
  if (invHb) delete invHb;
  if (matrix1) delete matrix1;
  if (matrix2) delete matrix2;
  if (matrix3) delete matrix3;
  if (matrix4) delete matrix4;
  if (matrix5) delete matrix5;
  if (vector1) delete vector1;

  time = clock.Check();
  PRINT("Running time: %g\n",time);
  FLUSH;
}

// Parse the inputs off the command line.
void ParseInputs(int argc, char *argv[]) {
  int i;

  // Set default values 
  gNU = 1;
  gSETNU = FALSE;
  gDATATYPE[0]='d';
  gHEADER = FALSE;
  gCONVERGENCE = TRUE;
  gPERTURB = 0;
  gNORMALIZE = FALSE;
  gMAXITER = 1000;
  gTOL = 1e-3;
  gPROMPT = TRUE;
  strcpy(gTESTFILENAME,"__none__");

  // Parse the inputs
  for (i=1; i < argc && (argv[i][0]=='-'); i++) {
    
    if (!strcmp(argv[i],"-nu")) {
      i++;
      gNU = atof(argv[i]);
      gSETNU = TRUE;
    } else if (!strcmp(argv[i],"-precision")) {
      i++;
      strcpy(gDATATYPE,argv[i]);
    } else if (!strcmp(argv[i],"-header"))
      gHEADER = TRUE;
    else if (!strcmp(argv[i],"-normalize"))
      gNORMALIZE = TRUE;
    else if (!strcmp(argv[i],"-noconverge")) 
      gCONVERGENCE = FALSE;
    else if (!strcmp(argv[i],"-prompt")) 
      gPROMPT = FALSE;
    else if (!strcmp(argv[i],"-perturb")) {
      i++;
      gPERTURB = atof(argv[i]);
    } else if (!strcmp(argv[i],"-test")) {
      i++;
      strcpy(gTESTFILENAME,argv[i]);
    } else if (!strcmp(argv[i],"-maxiter")) {
      i++;
      gMAXITER = atoi(argv[i]);
    } else if (!strcmp(argv[i],"-tol")) {
      i++;
      gTOL = atof(argv[i]);
    } else
      ThrowIt("Invalid option.");

  } // end-for

  if (gNU < 0)
    ThrowIt("Error: Nu must be positive.");

  if (i != argc-2) {
    PRINT("Active Support Vector Machine Classification Software.\n");
    PRINT("Copyright (C) 2000 David R. Musicant and Olvi L. Mangasarian.\n");
    PRINT(VERSION);
    PRINT("\n\n");
    PRINT("Syntax: asvm [options] training_file solution_file\n");
    PRINT("Possible choices are:\n");
    PRINT("-noconverge          Do not enforce convergence conditions\n");
    PRINT("-maxiter x           Specify maximum number of iterations as x\n");
    PRINT("-normalize           Normalize the data\n");
    PRINT("-nu                  Specify value x for nu\n");
    PRINT("-perturb x           Perturb the data by random values with max x\n");
    PRINT("-test x              Show performance on testing set x\n");
    PRINT("-precision double    Store data matrix as double precision\n");
    PRINT("-precision int       Store data matrix as integer values\n");
    PRINT("-precision short     Store data matrix as short values\n");
    PRINT("-noprompt            Do not prompt to stop if ASVM seems stuck\n");
    PRINT("-tol x               Specify convergence tolerance as x\n");
    exit(2);
  }

  strcpy(gSOURCEFILENAME,argv[i]);
  strcpy(gOUTPUTFILENAME,argv[i+1]);
}

// Dump the output from training to a flat file
void WriteWGamma(DenseVector *w,double gamma,double *mean, double *std,
		 char *outputFile) {
  int n = w->Size();
  FILE *out = fopen(outputFile,"w");
  if (!out)
    ThrowIt("Unable to write to output file.");
  fprintf(out,"%d\n%30.15g\n",n,gamma);
  int i;
  for (i=0; i < n; i++)
    fprintf(out,"%30.15g\n",(*w)[i]);
  for (i=0; i < n; i++)
    fprintf(out,"%30.15g\n",mean[i]);
  for (i=0; i < n; i++)
    fprintf(out,"%30.15f\n",std[i]);
  fclose(out);
}

// Read in the data, prepare for training, and perform testing
int main(int argc, char *argv[]) {
  
  try {
    ParseInputs(argc,argv);
    MatrixType *H;
    DenseVector *u;

    PRINT("Reading data from disk.\n");
    if (gDATATYPE[0]=='i' || gDATATYPE[0]=='I') {
      PRINT("Data storage type = int, ");
      H = new DenseMatrixType<int>(gSOURCEFILENAME,gHEADER,gPERTURB);
    } else if (gDATATYPE[0]=='s' || gDATATYPE[0]=='S') {
      PRINT("Data storage type = short, ");
      H = new DenseMatrixType<short>(gSOURCEFILENAME,gHEADER,gPERTURB);
    } else {
      PRINT("Data storage type = double, ");
      H = new DenseMatrixType<double>(gSOURCEFILENAME,gHEADER,gPERTURB);
    }
    FLUSH;

    // Pick a default value of nu of 1/size(H,1) if it has not been
    // picked already
    if (!gSETNU)
      gNU = 1./(H->rows);
    PRINT("nu = %g\n",gNU);

    // Handle normalization, if selected
    double *mean = new double[H->cols-1];
    double *std = new double[H->cols-1];
    if (gNORMALIZE) {
      H->GetMeanStd(mean,std);
      H->Normalize(mean,std);
    } else
      for (int i=0; i < H->cols-1; i++) {
	mean[i] = 0;
	std[i] = 1;
      }

    double obj, time;
    u = new DenseVector(H->rows);
    int numIter;
    PRINT("Beginning algorithm.\n");
    FLUSH;
    DenseMatrixType<double> *Hdouble;
    DenseMatrixType<int> *Hint;
    DenseMatrixType<short> *Hshort;
    if ((Hdouble=dynamic_cast<DenseMatrixType<double>*>(H)))
      AsvmTrain(Hdouble,gNU,gTOL,gMAXITER,u,obj,time,numIter,gCONVERGENCE,
		gPROMPT);
    else if ((Hint=dynamic_cast<DenseMatrixType<int>*>(H)))
      AsvmTrain(Hint,gNU,gTOL,gMAXITER,u,obj,time,numIter,gCONVERGENCE,
		gPROMPT);
    else if ((Hshort=dynamic_cast<DenseMatrixType<short>*>(H)))
      AsvmTrain(Hshort,gNU,gTOL,gMAXITER,u,obj,time,numIter,gCONVERGENCE,
		gPROMPT);
    else
      ThrowIt("Resolution of H failed.");

    // Determine w, gamma, accuracy
    DenseVector *w = new DenseVector((H->cols)-1);
    double gamma;
    H->ComputeWGamma(u,w,gamma);

    // Check for w=0, gamma=0 solution
    bool zeroSolution = TRUE;
    for (int i=0; i < w->Size() && zeroSolution; i++)
      if (((*w)[i]) != 0)
	zeroSolution = FALSE;
    if (gamma != 0) zeroSolution = FALSE;

    if (zeroSolution) {
      // Perturb H (which is A) somewhat
      H->Perturb(PERTURB_VALUE);
      if ((Hdouble=dynamic_cast<DenseMatrixType<double>*>(H))) {
	PRINT("Degenerate solution obtained. Perturbing and rerunning.\n");
	AsvmTrain(Hdouble,gNU,gTOL,gMAXITER,u,obj,time,numIter,
		      gCONVERGENCE,gPROMPT);
	H->ComputeWGamma(u,w,gamma);

	// Check again for w=0, gamma=0 solution
	bool zeroSolution = TRUE;
	for (int i=0; i < w->Size() && zeroSolution; i++)
	  if (((*w)[i]) != 0)
	    zeroSolution = FALSE;
	if (gamma != 0) zeroSolution = FALSE;

      } else if ((Hint=dynamic_cast<DenseMatrixType<int>*>(H))) {
	PRINT("Degenerate solution obtained. ");
	PRINT("Cannot perturb, since data stored as integers.\n");
      } else if ((Hshort=dynamic_cast<DenseMatrixType<short>*>(H))) {
	PRINT("Degenerate solution obtained. ");
	PRINT("Cannot perturb, since data stored as integers.\n");
      } else
	ThrowIt("Resolution of H failed.");
    }

    // If classification less than 40% due to pathological cases,
    // reverse classifier
    double accuracy = H->Accuracy(w,gamma);
    if (accuracy < 0.4 && !zeroSolution) {
      PRINT("Reversing classifier due to poor training accuracy.\n");
      for (int i=0; i < w->Size(); i++)
	w->dataMatrix[i] = -(w->dataMatrix[i]);
      gamma = -gamma;
    }
    PRINT("Train accuracy = %g\n",H->Accuracy(w,gamma));
    WriteWGamma(w,gamma,mean,std,gOUTPUTFILENAME);
    
    MatrixType *testH;
    if (strcmp(gTESTFILENAME,"__none__")) {
      if (gDATATYPE[0]=='i' || gDATATYPE[0]=='I')
	testH =  new DenseMatrixType<int>(gTESTFILENAME,gHEADER,gPERTURB);
      else if (gDATATYPE[0]=='s' || gDATATYPE[0]=='S')
	testH =  new DenseMatrixType<short>(gTESTFILENAME,gHEADER,gPERTURB);
      else
	testH =  new DenseMatrixType<double>(gTESTFILENAME,gHEADER,gPERTURB);
      if (gNORMALIZE)
	testH->Normalize(mean,std);
      PRINT("Test accuracy = %g\n",testH->Accuracy(w,gamma));
      delete(testH);
    }

    delete [] mean;
    delete [] std;
    delete(w);
    delete(H);
    delete(u);
    
  }
  catch(const char* text) {
    PRINT("%s\n",text);
  }

  return 0;
}

