///////////////////////////////////////////////////////////////////////////
// ASVM: Active Support Vector Machine                                   //
// Copyright 2000 David R. Musicant and Olvi L. Mangasarian.             //
// This software is free for academic use only.                          //
// For commercial use, contact musicant@cs.wisc.edu.                     //
///////////////////////////////////////////////////////////////////////////
// assvmmex.cpp: Mex file for linking MATLAB with ASVM method.           //
///////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <math.h>
#include "mex.h"
#include "matrix.h"
#include "AsvmTrain.cpp"
#include "Version.H"

#define PRINT mexPrintf
#define FLUSH
#define TRUE 1

// Wrapper class for MATLAB matrices
class MatlabInputMatrix {
 private:
  double *matrix;
 public:
  int m,n;

  MatlabInputMatrix(const mxArray *pointer) {
    matrix = (double *)mxGetPr(pointer);
    m = mxGetM(pointer);
    n = mxGetN(pointer);
  }

  inline double operator()(int i, int j) {
    return matrix[i+m*j];
  }
};

// Wrapper class for MATLAB matrices
class MatlabOutputMatrix {
private:
  double *matrix;
 public:
  int m,n;

  MatlabOutputMatrix(mxArray **pointer, int rows, int cols) {
    m = rows;
    n = cols;
    *pointer = mxCreateDoubleMatrix(m,n,mxREAL);
    matrix = (double *)mxGetPr(*pointer);
  }

  inline double operator()(int i, int j) {
    return matrix[i+m*j];
  }

  inline void Assign(int i,int j, double value) {
    matrix[i+m*j] = value;
  }
};

// Exceptions do not work as expected in a mexfile, so instead just
// spit out an error message
void ThrowIt(const char *text) {
  mexErrMsgTxt(text);
}

// Input arguments
#define	H_IN       prhs[0]
#define NU_IN      prhs[1]
#define TOL_IN     prhs[2]
#define MAXITER_IN prhs[3]
#define CONV_IN    prhs[4]
#define NORM_IN    prhs[5]
#define PROMPT_IN  prhs[6]

// Output aguments
#define W_OUT       plhs[0]
#define GAMMA_OUT   plhs[1]
#define OBJ_OUT     plhs[2]
#define TIME_OUT    plhs[3]
#define ITER_OUT    plhs[4]

void mexFunction(int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[])
{
  // Check for proper number of arguments
  if (nrhs != 7) {
    mexErrMsgTxt("Improper number of input arguments.");
  }
  
  // Input variables
  MatlabInputMatrix HMatlab(H_IN);
  double nu = mxGetScalar(NU_IN);
  double tol = mxGetScalar(TOL_IN);
  int maxIter = (int)mxGetScalar(MAXITER_IN);
  int converge = (int)mxGetScalar(CONV_IN);
  int normalize = (int)mxGetScalar(NORM_IN);
  int prompt = (int)mxGetScalar(PROMPT_IN);
  double *matH = (double *)mxGetPr(H_IN);

  // Output variables
  int m = HMatlab.m;
  int n = HMatlab.n;
  W_OUT = mxCreateDoubleMatrix(n-1,1,mxREAL);
  GAMMA_OUT = mxCreateDoubleMatrix(1,1,mxREAL);
  OBJ_OUT = mxCreateDoubleMatrix(1,1,mxREAL);
  TIME_OUT = mxCreateDoubleMatrix(1,1,mxREAL);
  ITER_OUT = mxCreateDoubleMatrix(1,1,mxREAL);
  double *wOut = (double *)mxGetPr(W_OUT);
  double *gamma = (double *)mxGetPr(GAMMA_OUT);
  double *obj = (double *)mxGetPr(OBJ_OUT);
  double *time = (double *)mxGetPr(TIME_OUT);
  double *numIter = (double *)mxGetPr(ITER_OUT);

  // Convert the MATLAB style H matrix to a C++ style H matrix
  DenseMatrixType<double> *H = new DenseMatrixType<double>(m,n);
  int i,j;
  for (i=0; i < m; i++)
    for (j=0; j < n; j++)
      H->dataMatrix[i][j] = HMatlab(i,j);
  // Normalize if needed
  if (normalize) {
    double *mean = new double[H->cols-1];
    double *std = new double[H->cols-1];
    H->GetMeanStd(mean,std);
    H->Normalize(mean,std);
  }

  DenseVector *u = new DenseVector(H->rows);
  DenseVector *w = new DenseVector(H->cols - 1);
  int numIterReturn;

  AsvmTrain(H,nu,tol,maxIter,u,*obj,*time,numIterReturn,converge,prompt);
  *numIter = numIterReturn;

  H->ComputeWGamma(u,w,*gamma);

  // Check for w=0, gamma=0 solution
  bool zeroSolution = TRUE;
  for (i=0; i < w->Size() && zeroSolution; i++)
    if (((*w)[i]) != 0)
      zeroSolution = FALSE;
  if (*gamma != 0) zeroSolution = FALSE;

  // If classification less than 40% due to pathological cases,
  // reverse classifier
  if (!zeroSolution) {
    double accuracy = H->Accuracy(w,*gamma);
    if (accuracy < 0.4) {
      PRINT("Reversing classifier due to poor training accuracy.\n");
      for (i=0; i < w->Size(); i++)
	w->dataMatrix[i] = -(w->dataMatrix[i]);
      *gamma = -*gamma;
    }
    PRINT("Train accuracy = %g\n",H->Accuracy(w,*gamma));
  }

  // Copy in w
  for (i=0; i < w->Size(); i++)
    wOut[i] = (*w)[i];

  delete H;
  delete u;

  return;
}
